def wildcard_match(pattern, text, case_insensitive=True):
    """
    Performs a case-insensitive wildcard match against two strings.
    This method works with pseduo-regex chars; specifically ? and * are supported.
    An asterisk (*) represents any combination of characters.
    A question mark (?) represents any single character.
    :param str pattern: the regex-like pattern to be compared against
    :param str text: the string to compare against the pattern
    :param boolean case_insensitive: dafault is True
    return whether the text matches the pattern
    """
    if pattern is None or text is None:
        return False

    if len(pattern) == 0:
        return len(text) == 0

    # Check the special case of a single * pattern, as it's common
    if pattern == '*':
        return True

    # If elif logic Checking different conditions like match between the first i chars in text
    # and the first p chars in pattern, checking pattern has '?' or '*' also check for case_insensitivity
    # iStar is introduced to store length of the text and i, p and pStar for indexing
    i = 0
    p = 0
    iStar = len(text)
    pStar = 0
    while i < len(text):
        if p < len(pattern) and text[i] == pattern[p]:
            i = i + 1
            p = p + 1

        elif p < len(pattern) and case_insensitive and text[i].lower() == pattern[p].lower():
            i = i + 1
            p = p + 1

        elif p < len(pattern) and pattern[p] == '?':
            i = i + 1
            p = p + 1

        elif p < len(pattern) and pattern[p] == '*':
            iStar = i
            pStar = p
            p += 1

        elif iStar != len(text):
            iStar += 1
            i = iStar
            p = pStar + 1

        else:
            return False

    while p < len(pattern) and pattern[p] == '*':
        p = p + 1

    return p == len(pattern) and i == len(text)



import sys
import traceback


def get_stacktrace(limit=None):
    """
    Get a full stacktrace for the current state of execution.

    Include the current state of the stack, minus this function.
    If there is an active exception, include the stacktrace information from
    the exception as well.

    :param int limit:
        Optionally limit stack trace size results. This parmaeters has the same
        meaning as the `limit` parameter in `traceback.print_stack`.
    :returns:
        List of stack trace objects, in the same form as
        `traceback.extract_stack`.
    """
    if limit is not None and limit == 0:
        # Nothing to return. This is consistent with the behavior of the
        # functions in the `traceback` module.
        return []

    stack = traceback.extract_stack()
    # Remove this `get_stacktrace()` function call from the stack info.
    # For what we want to report, this is superfluous information and arguably
    # adds garbage to the report.
    # Also drop the `traceback.extract_stack()` call above from the returned
    # stack info, since this is also superfluous.
    stack = stack[:-2]

    _exc_type, _exc, exc_traceback = sys.exc_info()
    if exc_traceback is not None:
        # If and only if there is a currently triggered exception, combine the
        # exception traceback information with the current stack state to get a
        # complete trace.
        exc_stack = traceback.extract_tb(exc_traceback)
        stack += exc_stack

    # Limit the stack trace size, if a limit was specified:
    if limit is not None:
        # Copy the behavior of `traceback` functions with a `limit` argument.
        # See https://docs.python.org/3/library/traceback.html.
        if limit > 0:
            # limit > 0: include the last `limit` items
            stack = stack[-limit:]
        else:
            # limit < 0: include the first `abs(limit)` items
            stack = stack[:abs(limit)]
    return stack



import os
import logging
from distutils.util import strtobool

log = logging.getLogger(__name__)


class SDKConfig(object):
    """
    Global Configuration Class that defines SDK-level configuration properties.

    Enabling/Disabling the SDK:
        By default, the SDK is enabled unless if an environment variable AWS_XRAY_SDK_ENABLED
            is set. If it is set, it needs to be a valid string boolean, otherwise, it will default
            to true. If the environment variable is set, all calls to set_sdk_enabled() will
            prioritize the value of the environment variable.
        Disabling the SDK affects the recorder, patcher, and middlewares in the following ways:
        For the recorder, disabling automatically generates DummySegments for subsequent segments
            and DummySubsegments for subsegments created and thus not send any traces to the daemon.
        For the patcher, module patching will automatically be disabled. The SDK must be disabled
            before calling patcher.patch() method in order for this to function properly.
        For the middleware, no modification is made on them, but since the recorder automatically
            generates DummySegments for all subsequent calls, they will not generate segments/subsegments
            to be sent.

    Environment variables:
        "AWS_XRAY_SDK_ENABLED" - If set to 'false' disables the SDK and causes the explained above
            to occur.
    """
    XRAY_ENABLED_KEY = 'AWS_XRAY_SDK_ENABLED'
    DISABLED_ENTITY_NAME = 'dummy'

    __SDK_ENABLED = None

    @classmethod
    def __get_enabled_from_env(cls):
        """
        Searches for the environment variable to see if the SDK should be disabled.
        If no environment variable is found, it returns True by default.

        :return: bool - True if it is enabled, False otherwise.
        """
        env_var_str = os.getenv(cls.XRAY_ENABLED_KEY, 'true')
        try:
            return bool(strtobool(env_var_str))
        except ValueError:
            log.warning("Invalid literal passed into environment variable `AWS_XRAY_SDK_ENABLED`. Defaulting to True...")
            return True  # If an invalid parameter is passed in, we return True.

    @classmethod
    def sdk_enabled(cls):
        """
        Returns whether the SDK is enabled or not.
        """
        if cls.__SDK_ENABLED is None:
            cls.__SDK_ENABLED = cls.__get_enabled_from_env()
        return cls.__SDK_ENABLED

    @classmethod
    def set_sdk_enabled(cls, value):
        """
        Modifies the enabled flag if the "AWS_XRAY_SDK_ENABLED" environment variable is not set,
        otherwise, set the enabled flag to be equal to the environment variable. If the
        env variable is an invalid string boolean, it will default to true.

        :param bool value: Flag to set whether the SDK is enabled or disabled.

        Environment variables AWS_XRAY_SDK_ENABLED overrides argument value.
        """
        # Environment Variables take precedence over hardcoded configurations.
        if cls.XRAY_ENABLED_KEY in os.environ:
            cls.__SDK_ENABLED = cls.__get_enabled_from_env()
        else:
            if type(value) == bool:
                cls.__SDK_ENABLED = value
            else:
                cls.__SDK_ENABLED = True
                log.warning("Invalid parameter type passed into set_sdk_enabled(). Defaulting to True...")



import logging
from random import Random
import time
import threading

log = logging.getLogger(__name__)


class TargetPoller(object):
    """
    The poller to report the current statistics of all
    centralized sampling rules and retrieve the new allocated
    sampling quota and TTL from X-Ray service.
    """
    def __init__(self, cache, rule_poller, connector):
        self._cache = cache
        self._rule_poller = rule_poller
        self._connector = connector
        self._random = Random()
        self._interval = 10 # default 10 seconds interval on sampling targets fetch

    def start(self):
        poller_thread = threading.Thread(target=self._worker)
        poller_thread.daemon = True
        poller_thread.start()

    def _worker(self):
        while True:
            try:
                time.sleep(self._interval + self._get_jitter())
                self._do_work()
            except Exception:
                log.error("Encountered an issue while polling targets.", exc_info=True)

    def _do_work(self):
        candidates = self._get_candidates(self._cache.rules)
        if not candidates:
            log.debug('There is no sampling rule statistics to report. Skipping')
            return None
        targets, rule_freshness = self._connector.fetch_sampling_target(candidates)
        self._cache.load_targets(targets)

        if rule_freshness > self._cache.last_updated:
            log.info('Performing out-of-band sampling rule polling to fetch updated rules.')
            self._rule_poller.wake_up()

    def _get_candidates(self, all_rules):
        """
        Don't report a rule statistics if any of the conditions is met:
        1. The report time hasn't come(some rules might have larger report intervals).
        2. The rule is never matched.
        """
        candidates = []
        for rule in all_rules:
            if rule.ever_matched() and rule.time_to_report():
                candidates.append(rule)
        return candidates

    def _get_jitter(self):
        """
        A random jitter of up to 0.1 seconds is injected after every run
        to ensure all poller calls eventually get evenly distributed
        over the polling interval window.
        """
        return self._random.random() / self._interval



import logging

log = logging.getLogger(__name__)

ROOT = 'Root'
PARENT = 'Parent'
SAMPLE = 'Sampled'
SELF = 'Self'

HEADER_DELIMITER = ";"


class TraceHeader(object):
    """
    The sampling decision and trace ID are added to HTTP requests in
    tracing headers named ``X-Amzn-Trace-Id``. The first X-Ray-integrated
    service that the request hits adds a tracing header, which is read
    by the X-Ray SDK and included in the response. Learn more about
    `Tracing Header <http://docs.aws.amazon.com/xray/latest/devguide/xray-concepts.html#xray-concepts-tracingheader>`_.
    """
    def __init__(self, root=None, parent=None, sampled=None, data=None):
        """
        :param str root: trace id
        :param str parent: parent id
        :param int sampled: 0 means not sampled, 1 means sampled
        :param dict data: arbitrary data fields
        """
        self._root = root
        self._parent = parent
        self._sampled = None
        self._data = data

        if sampled is not None:
            if sampled == '?':
                self._sampled = sampled
            if sampled is True or sampled == '1' or sampled == 1:
                self._sampled = 1
            if sampled is False or sampled == '0' or sampled == 0:
                self._sampled = 0

    @classmethod
    def from_header_str(cls, header):
        """
        Create a TraceHeader object from a tracing header string
        extracted from a http request headers.
        """
        if not header:
            return cls()

        try:
            params = header.strip().split(HEADER_DELIMITER)
            header_dict = {}
            data = {}

            for param in params:
                entry = param.split('=')
                key = entry[0]
                if key in (ROOT, PARENT, SAMPLE):
                    header_dict[key] = entry[1]
                # Ignore any "Self=" trace ids injected from ALB.
                elif key != SELF:
                    data[key] = entry[1]

            return cls(
                root=header_dict.get(ROOT, None),
                parent=header_dict.get(PARENT, None),
                sampled=header_dict.get(SAMPLE, None),
                data=data,
            )

        except Exception:
            log.warning("malformed tracing header %s, ignore.", header)
            return cls()

    def to_header_str(self):
        """
        Convert to a tracing header string that can be injected to
        outgoing http request headers.
        """
        h_parts = []
        if self.root:
            h_parts.append(ROOT + '=' + self.root)
        if self.parent:
            h_parts.append(PARENT + '=' + self.parent)
        if self.sampled is not None:
            h_parts.append(SAMPLE + '=' + str(self.sampled))
        if self.data:
            for key in self.data:
                h_parts.append(key + '=' + self.data[key])

        return HEADER_DELIMITER.join(h_parts)

    @property
    def root(self):
        """
        Return trace id of the header
        """
        return self._root

    @property
    def parent(self):
        """
        Return the parent segment id in the header
        """
        return self._parent

    @property
    def sampled(self):
        """
        Return the sampling decision in the header.
        It's 0 or 1 or '?'.
        """
        return self._sampled

    @property
    def data(self):
        """
        Return the arbitrary fields in the trace header.
        """
        return self._data



import os

from django.conf import settings as django_settings
from django.test.signals import setting_changed

DEFAULTS = {
    'AWS_XRAY_DAEMON_ADDRESS': '127.0.0.1:2000',
    'AUTO_INSTRUMENT': True,
    'AWS_XRAY_CONTEXT_MISSING': 'RUNTIME_ERROR',
    'PLUGINS': (),
    'SAMPLING': True,
    'SAMPLING_RULES': None,
    'SAMPLER': None,
    'AWS_XRAY_TRACING_NAME': None,
    'DYNAMIC_NAMING': None,
    'STREAMING_THRESHOLD': None,
    'MAX_TRACE_BACK': None,
    'STREAM_SQL': True,
    'PATCH_MODULES': [],
    'AUTO_PATCH_PARENT_SEGMENT_NAME': None,
    'IGNORE_MODULE_PATTERNS': [],
}

XRAY_NAMESPACE = 'XRAY_RECORDER'

SUPPORTED_ENV_VARS = ('AWS_XRAY_DAEMON_ADDRESS',
                      'AWS_XRAY_CONTEXT_MISSING',
                      'AWS_XRAY_TRACING_NAME',
                      )


class XRaySettings(object):
    """
    A object of Django settings to easily modify certain fields.
    The precedence for configurations at different places is as follows:
    environment variables > user settings in settings.py > default settings
    """
    def __init__(self, user_settings=None):

        self.defaults = DEFAULTS

        if user_settings:
            self._user_settings = user_settings

    @property
    def user_settings(self):

        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(django_settings, XRAY_NAMESPACE, {})

        return self._user_settings

    def __getattr__(self, attr):

        if attr not in self.defaults:
            raise AttributeError('Invalid setting: %s' % attr)

        if self.user_settings.get(attr, None) is not None:
            if attr in SUPPORTED_ENV_VARS:
                return os.getenv(attr, self.user_settings[attr])
            else:
                return self.user_settings[attr]
        elif attr in SUPPORTED_ENV_VARS:
            return os.getenv(attr, self.defaults[attr])
        else:
            return self.defaults[attr]


settings = XRaySettings()


def reload_settings(*args, **kwargs):
    """
    Reload X-Ray user settings upon Django server hot restart
    """
    global settings
    setting, value = kwargs['setting'], kwargs['value']
    if setting == XRAY_NAMESPACE:
        settings = XRaySettings(value)


setting_changed.connect(reload_settings)
